/*-
 * ============LICENSE_START=======================================================
 *  Copyright (C) 2021-2023 Nordix Foundation.
 * ================================================================================
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 * SPDX-License-Identifier: Apache-2.0
 * ============LICENSE_END=========================================================
 */

package org.onap.policy.clamp.acm.participant.kubernetes.handler;

import java.io.IOException;
import java.lang.invoke.MethodHandles;
import java.util.HashMap;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import javax.ws.rs.core.Response;
import lombok.AccessLevel;
import lombok.Getter;
import lombok.Setter;
import org.onap.policy.clamp.acm.participant.intermediary.api.AutomationCompositionElementListener;
import org.onap.policy.clamp.acm.participant.intermediary.api.ParticipantIntermediaryApi;
import org.onap.policy.clamp.acm.participant.kubernetes.exception.ServiceException;
import org.onap.policy.clamp.acm.participant.kubernetes.helm.PodStatusValidator;
import org.onap.policy.clamp.acm.participant.kubernetes.models.ChartInfo;
import org.onap.policy.clamp.acm.participant.kubernetes.service.ChartService;
import org.onap.policy.clamp.models.acm.concepts.AcElementDeploy;
import org.onap.policy.clamp.models.acm.concepts.DeployState;
import org.onap.policy.common.utils.coder.Coder;
import org.onap.policy.common.utils.coder.CoderException;
import org.onap.policy.common.utils.coder.StandardCoder;
import org.onap.policy.models.base.PfModelException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * This class handles implementation of automationCompositionElement updates.
 */
@Component
public class AutomationCompositionElementHandler implements AutomationCompositionElementListener {
    private static final Logger LOGGER = LoggerFactory.getLogger(MethodHandles.lookup().lookupClass());

    private ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());

    // Map of helm installation and the status of corresponding pods
    @Getter
    private static Map<String, Map<String, String>> podStatusMap = new ConcurrentHashMap<>();
    private static final Coder CODER = new StandardCoder();

    @Autowired
    private ChartService chartService;

    @Setter
    private ParticipantIntermediaryApi intermediaryApi;

    // Map of acElement Id and installed Helm charts
    @Getter(AccessLevel.PACKAGE)
    private final Map<UUID, ChartInfo> chartMap = new HashMap<>();

    // Default thread config values
    private static class ThreadConfig {
        private int uninitializedToPassiveTimeout = 60;
        private int podStatusCheckInterval = 30;
    }

    /**
     * Callback method to handle a automation composition element state change.
     *
     * @param automationCompositionElementId the ID of the automation composition element
     */
    @Override
    public synchronized void undeploy(UUID automationCompositionId, UUID automationCompositionElementId) {
        var chart = chartMap.get(automationCompositionElementId);
        if (chart != null) {
            LOGGER.info("Helm deployment to be deleted {} ", chart.getReleaseName());
            try {
                chartService.uninstallChart(chart);
                intermediaryApi.updateAutomationCompositionElementState(automationCompositionId,
                        automationCompositionElementId, DeployState.UNDEPLOYED, null, "Undeployed");
                chartMap.remove(automationCompositionElementId);
                podStatusMap.remove(chart.getReleaseName());
            } catch (ServiceException se) {
                LOGGER.warn("Deletion of Helm deployment failed", se);
            }
        }
    }

    /**
     * Callback method to handle an update on a automation composition element.
     *
     * @param automationCompositionId the automationComposition Id
     * @param element the information on the automation composition element
     * @param properties properties Map
     * @throws PfModelException in case of an exception
     */
    @Override
    public synchronized void deploy(UUID automationCompositionId, AcElementDeploy element,
            Map<String, Object> properties) throws PfModelException {
        @SuppressWarnings("unchecked")
        var chartData = (Map<String, Object>) properties.get("chart");

        LOGGER.info("Installation request received for the Helm Chart {} ", chartData);
        try {
            var chartInfo = CODER.convert(chartData, ChartInfo.class);
            if (chartService.installChart(chartInfo)) {
                chartMap.put(element.getId(), chartInfo);

                var config = CODER.convert(properties, ThreadConfig.class);
                checkPodStatus(automationCompositionId, element.getId(), chartInfo,
                        config.uninitializedToPassiveTimeout, config.podStatusCheckInterval);
            }
        } catch (ServiceException | CoderException | IOException e) {
            LOGGER.warn("Installation of Helm chart failed", e);
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            throw new PfModelException(Response.Status.BAD_REQUEST, "Error invoking ExecutorService ", e);
        } catch (ExecutionException e) {
            throw new PfModelException(Response.Status.BAD_REQUEST, "Error retrieving pod status result ", e);
        }
    }

    /**
     * Invoke a new thread to check the status of deployed pods.
     *
     * @param chart ChartInfo
     */
    public void checkPodStatus(UUID automationCompositionId, UUID elementId, ChartInfo chart, int timeout,
            int podStatusCheckInterval) throws ExecutionException, InterruptedException {
        // Invoke runnable thread to check pod status
        var result = executor.submit(new PodStatusValidator(chart, timeout, podStatusCheckInterval), "Done");
        if (!result.get().isEmpty()) {
            LOGGER.info("Pod Status Validator Completed: {}", result.isDone());
            intermediaryApi.updateAutomationCompositionElementState(automationCompositionId, elementId,
                    DeployState.DEPLOYED, null, "Deployed");
        }
    }
}
